unit PalDemo;

{ Demostration of palettes and animation in a Delphi component.
  Copyright  1997 Tempest Software, Inc.

  The TPaletteDemo component displays a color gradient, with
  a user-specified starting and ending color, and number of
  color steps between them. The gradient works on all systems.

  The Animate procedure randomly shifts the starting and ending
  colors, and animates the palette to change the display of the gradient
  without repainting. Palette animation works only on displays that
  use a palette, namely, 256-color video adapters.
}

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs;

type
  TPaletteDemo = class(TCustomControl)
  private
    { Colors for the end pointersin the gradient }
    fColorTop, fColorBottom: TColor;
    { The same colors after mutation. }
    fColTop, fColBottom: TColor;
    { Delta for changing the color each iteration }
    DeltaTop, DeltaBottom: TColor;
    { Number of steps in the gradient }
    fNumSteps: Byte;
    { Palette information }
    PalHandle: HPalette;      { handle to the palette }
    LogPal: PLogPalette;      { pointer to logical palette }
    PalSize: Integer;         { size of LogPal memory }
    procedure SetColorTop(Value: TColor);
    procedure SetColorBottom(Value: TColor);
    procedure SetNumSteps(Value: Byte);
  protected
    { Delphi calls GetPalette for every control. Return the palette
      handle to let Delphi take care of selecting and realizing that
      palette when it paints this control. }
    function GetPalette: HPalette; override;
    { Paint this control by drawing the gradient. }
    procedure Paint; override;
    { Create the gradient palette and return its handle. }
    function MakePalette: HPalette;
    { Randomly mutate the colors and change the palette entries. }
    procedure MutateColors;
    { Destroy the palette after the control is done using it. }
    procedure DestroyPalette;
    { Change the colors in the new palette. }
    procedure UseNewPalette;
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    { Mutate the colors and animate the system palette for those new colors. }
    procedure Animate;
    { Return the control's canvas. }
    property Canvas;
  published
    property OnMouseMove;
    property OnMouseDown;
    property OnMouseUp;
    property OnExit;
    property OnEnter;
    property OnClick;
    property OnDblClick;
    property Align;
    property Height default 100;
    property Width default 100;
    property ColorTop: TColor read fColorTop write SetColorTop default clBlue;
    property ColorBottom: TColor read fColorBottom write SetColorBottom default clBlack;
    property NumSteps: Byte read fNumSteps write SetNumSteps default 64;
  end;

procedure Register;

implementation

{ Initialize the gradient palette entries. Build the palette entries
  so they smoothly progress from the bottom color to the top color. }
procedure MakeGradient(Top, Bottom: TColor; var Pal: TLogPalette);
var
  I: Integer;
begin
{$R-}{ Turn off range checking to access palette entries. }
  for I := 0 to Pal.palNumEntries-1 do
  begin
    Pal.palPalEntry[I].peRed   :=
      MulDiv(Pal.palNumEntries-I-1, GetRValue(Bottom), Pal.palNumEntries-1) +
      MulDiv(I, GetRValue(Top), Pal.palNumEntries-1);

    Pal.palPalEntry[I].peGreen :=
      MulDiv(Pal.palNumEntries-I-1, GetGValue(Bottom), Pal.palNumEntries-1) +
      MulDiv(I, GetGValue(Top), Pal.palNumEntries-1);

    Pal.palPalEntry[I].peBlue  :=
      MulDiv(Pal.palNumEntries-I-1, GetBValue(Bottom), Pal.palNumEntries-1) +
      MulDiv(I, GetBValue(Top), Pal.palNumEntries-1);

    { To animate the palette, the palette entry must have the Pc_Reserved flag. }
    Pal.palPalEntry[I].peFlags := PC_Reserved;
  end;
{$R+}
end;

{ Make a gradient palette and return the palette handle. Call this once
  when initializing the control. After that, just change the palette
  entries without making a new palette. If the user changes the number
  of steps, then recreate the palette with a new size. }
function TPaletteDemo.MakePalette: HPalette;
begin
  { TLogPalette already has one palette entry, so we need NumStep-1 more. }
  PalSize := SizeOf(TLogPalette) + (NumSteps-1)*SizeOf(TPaletteEntry);
  GetMem(LogPal, PalSize);
  LogPal^.palVersion := $300;
  LogPal^.palNumEntries := NumSteps;
  MakeGradient(ColorToRgb(fColTop), ColorToRgb(fColBottom), LogPal^);
  Result := CreatePalette(LogPal^);
end;

{ Create and initialize the palette demo control. Initialize the
  color deltas to random values, but make sure they are at least 1,
  so the colors really animate. Zero would be boring. Start with
  64 steps because most palette devices use 18 bits per pixel,
  which means 6 bits per color, or 64 distinct colors. }
constructor TPaletteDemo.Create(Owner: TComponent);
begin
  inherited Create(Owner);

  fColorTop := clBlue;
  fColTop := fColorTop;
  fColorBottom := clBlack;
  fColBottom := fColBottom;
  fNumSteps := 64;
  DeltaTop := Rgb(1 + Random(32), 1 + Random(32), 1 + Random(32));
  DeltaBottom := Rgb(1 + Random(32), 1 + Random(32), 1 + Random(32));

  Height := 100;
  Width := 100;
end;

{ Destroy the control. Clean up by freeing the palette resource and memory. }
destructor TPaletteDemo.Destroy;
begin
  DestroyPalette;
  inherited Destroy;
end;

{ Destroy the palette and free the palette memory. }
procedure TPaletteDemo.DestroyPalette;
begin
  if PalHandle <> 0 then
  begin
    DeleteObject(PalHandle);
    PalHandle := 0;
  end;
  FreeMem(LogPal, PalSize);
  LogPal := nil;
end;

{ Change the top color. }
procedure TPaletteDemo.SetColorTop(Value: TColor);
begin
  if Value <> ColorTop then
  begin
    fColorTop := Value;
    fColTop := fColorTop;
    UseNewPalette;
    Invalidate;
  end;
end;

{ Change the bottom color. }
procedure TPaletteDemo.SetColorBottom(Value: TColor);
begin
  if Value <> ColorBottom then
  begin
    fColorBottom := Value;
    fColBottom := fColorBottom;
    UseNewPalette;
    Invalidate;
  end;
end;

{ Change the number of steps in the gradient. Note that
  this requires rebuilding the palette with a new size. }
procedure TPaletteDemo.SetNumSteps(Value: Byte);
begin
  if Value <> NumSteps then
  begin
    fNumSteps := Value;
    if fNumSteps = 0 then
      fNumSteps := 1;
    DestroyPalette;
    Palhandle := MakePalette;
    Invalidate;
  end;
end;

{ Return the palette handle. }
function TPaletteDemo.GetPalette: HPalette;
begin
  if PalHandle = 0 then
    PalHandle := MakePalette;
  Result := PalHandle;
end;

{ Change the colors to animate the palette. }
procedure TPaletteDemo.MutateColors;
  { Mutate one component of a color. }
  procedure MutateOne(var Color: Byte; var Delta: ShortInt);
  begin
    if Integer(Color) + Delta < 0 then
    begin
      Color := 0;
      Delta := 1 + Random(32);
    end
    else if Integer(Color) + Delta > 255 then
    begin
      Color := 255;
      Delta := -Random(32) - 1;
    end
    else
      Inc(Color, Delta);
  end;

  { Mutate all components in a color. }
  procedure MutateAll(var Color, Delta: TColor);
  var
    R, G, B: Byte;
    DR, DG, DB: ShortInt;
  begin
    Color := ColorToRgb(Color);
    R := GetRValue(Color);
    G := GetGValue(Color);
    B := GetBValue(Color);
    { Delta cannot be a system color, so no need for ColorToRgb }
    DR := ShortInt(GetRValue(Delta));
    DG := ShortInt(GetGValue(Delta));
    DB := ShortInt(GetBValue(Delta));
    MutateOne(R, DR);
    MutateOne(G, DG);
    MutateOne(B, DB);
    Color := RGB(R, G, B);
    Delta := RGB(Byte(DR), Byte(DG), Byte(DB));
  end;
begin
  MutateAll(fColTop, DeltaTop);
  MutateAll(fColBottom, DeltaBottom);
end;

{ Paint the gradient. Tell Windows to use the palette when
  painting this control, even if the display does not use
  a system palette. To do this, we must explicitly select
  the palette and use PaletteIndex as the color.

  To paint the gradient, divide the canvas into NumSteps bands,
  and fill each band with a solid color, chosen from the palette. }
procedure TPaletteDemo.Paint;
var
  I: Integer;
  Y: Integer;          { vertical position on canvas }
  Rect: TRect;         { rectangle for filling one band in the gradient }
  OldPal: HPalette;    { old palette }
begin
  Rect := ClientRect;
  Y := 0;
  for I := 0 to NumSteps-1 do
  begin
    with Canvas do
    begin
      OldPal := SelectPalette(Handle, GetPalette, False);
      try
      {$R-}
        with LogPal^.palPalEntry[I] do
          Brush.Color := PaletteIndex(I);
      {$R+}
        Rect.Top := Y;
        Y := MulDiv(Height, I + 1, NumSteps);
        Rect.Bottom := Y;
        FillRect(Rect);
      finally
        SelectPalette(Handle, OldPal, False);
      end;
    end;
  end;
end;

{ Animate the control by mutating the color and then telling
  Windows to use the new colors. }
procedure TPaletteDemo.Animate;
begin
  MutateColors;
  UseNewPalette;
end;

{ Update the palette and tell Windows to use that new palette
  in the control. You must select the palette before animating it.
  Note that AnimatePalette does nothing on systems that do not
  use a display palette (that is, <> 256 colors). }
procedure TPaletteDemo.UseNewPalette;
var
  OldPal: HPalette;
begin
  GetPalette; { Make sure the palette has been created. }
  MakeGradient(ColorToRgb(fColTop), ColorToRgb(fColBottom), LogPal^);
  OldPal := SelectPalette(Canvas.Handle, GetPalette, False);
  try
  { Deal with gratuitous differences between versions of Delphi. Sigh... }
  {$ifdef WIN32}
    AnimatePalette(GetPalette, 0, LogPal^.palNumEntries, @LogPal^.palPalEntry[0]);
  {$else}
    AnimatePalette(GetPalette, 0, LogPal^.palNumEntries, LogPal^.palPalEntry);
  {$endif}
  finally
    SelectPalette(Canvas.Handle, OldPal, False);
  end;
end;

procedure Register;
begin
  RegisterComponents('Tempest', [TPaletteDemo]);
end;

end.
